package sillysagiri;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;

import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.Sprite;

import sillysagiri.scene.DeathError;

public class BMF {
  public static final byte ID = 0;
  public static final byte X = 1;
  public static final byte Y = 2;
  public static final byte WIDTH = 3;
  public static final byte HEIGHT = 4;
  public static final byte XOFF = 5;
  public static final byte YOFF = 6;
  public static final byte XADV = 7;

  public Image img;
  public short fontSize;
  public short lineHeight;
  public short base;
  public short[][] chars;

  public BMF()
  {
  }

  public static final BMF Create(String path_img, String path_fnt) throws Exception
  {
    BMF bmf = new BMF();
    bmf.img = Image.createImage(path_img);
    bmf.Parse_fnt(path_fnt);

    return bmf;
  }

  public static final BMF Create(Image img, String path_fnt) throws Exception
  {
    BMF bmf = new BMF();
    bmf.img = img;
    bmf.Parse_fnt(path_fnt);

    return bmf;
  }

  protected void Parse_fnt(String path_fnt) throws Exception
  {
    InputStream in = this.getClass().getResourceAsStream(path_fnt);
    byte[] header = new byte[4];

    int size_read = in.read(header, 0, 4);
    if (!(size_read == 4) && 
        !(header[0] != 'B') && 
        !(header[1] != 'M') &&
        !(header[2] != 'F') &&
        !(header[3] != 3))
    {
      throw new Exception("File is not bmf " + path_fnt);
    }
    
    header = null;
    byte[] buffer = new byte[20];
    while(size_read != -1)
    {
      int length;
      int type;

      // read type & length
      size_read = in.read(buffer, 0, 5);
      if (size_read == -1) break;

      // get type from buffer
      type = buffer[0];

      // convert byte into int;
      length = ((buffer[4] & 0xFF) << 24) |
               ((buffer[3] & 0xFF) << 16) |
               ((buffer[2] & 0xFF) << 8)  |
               (buffer[1] & 0xFF);
      
      switch (type) {
        case 1: // info block
        {
          // font size is signed
          size_read = in.read(buffer, 0, 2);
          fontSize = (short)( ((buffer[1]) << 8) |
                              (buffer[0]) );
          
          // skip the rest
          in.skip(length - 2);
        }
        break;
        
        case 2: // common block
        {
          size_read = in.read(buffer, 0, 4);
          lineHeight = (short)( ((buffer[1] & 0xFF) << 8) |
                                (buffer[0]) & 0xFF);

          base = (short)( ((buffer[3] & 0xFF) << 8) |
                          (buffer[2]) & 0xFF);

          // skip the rest
          in.skip(length - 4);
        }
        break;

        case 3: // pages block
        {
          // skip pages
          in.skip(length);
        }
        break;

        case 4: // chars block
        {
          chars = new short[96][8];
          int count = length/20;
          for (int i=0; i<count; i++)
          {
            size_read = in.read(buffer, 0, 20);
            int id = ((buffer[3] & 0xFF) << 24) |
                     ((buffer[2] & 0xFF) << 16) |
                     ((buffer[1] & 0xFF) << 8) |
                     (buffer[0] & 0xFF);

            // TODO: handle more glyph
            if (id < 32 || id > 128) continue;

            chars[id-32][ID] = (short)(((buffer[1] & 0xFF) << 8) | (buffer[0] & 0xFF));
            chars[id-32][X] = (short)(((buffer[5] & 0xFF) << 8) | (buffer[4] & 0xFF));
            chars[id-32][Y] = (short)(((buffer[7] & 0xFF) << 8) | (buffer[6] & 0xFF));
            chars[id-32][WIDTH] = (short)(((buffer[9] & 0xFF) << 8) | (buffer[8] & 0xFF));
            chars[id-32][HEIGHT] = (short)(((buffer[11] & 0xFF) << 8) | (buffer[10] & 0xFF));
            chars[id-32][XOFF] = (short)((buffer[13] << 8) | buffer[12]);
            chars[id-32][YOFF] = (short)((buffer[15] << 8) | buffer[14]);
            chars[id-32][XADV] = (short)((buffer[17] << 8) | buffer[16]);
          }
        }
        break;

        default:
          break;
      }
    }
  }

  /* 
   * this is cpu intensive process
   * consider creating new BMF for each color instead
   * 
   * limitation:
   * currently can only be used once
   * since its replacing white pixel with tinted color
   */
  public void SetColor(int rgb)
  {
    // int[] buffer = new int[img.getWidth()*img.getHeight()];
    // img.getRGB(buffer, 0, img.getWidth(), 0, 0, img.getWidth(), img.getHeight());
    
    // for (int i=0; i<buffer.length; i++)
    // {
      //   if (buffer[i] == 0xFF000000) buffer[i] = rgb;
      //   else buffer[i] = 0x00000000;
      // }

    try 
    {
      // ByteArrayOutputStream out = new ByteArrayOutputStream(128*64);
      // Gzip.gunzip(this.getClass().getResourceAsStream("/silly/bmf/nds12_0.silly.gz"), out);

      // GZIPInputStream in = new GZIPInputStream(this.getClass().getResourceAsStream("/silly/bmf/nds12_0.silly.gz"));

      InputStream in = this.getClass().getResourceAsStream("/silly/bmf/nds12_0.silly");
      int[] buffer = new int[128*64];
      
      // read header
      byte[] size = new byte[4];
      in.read(size, 0, 4);
      int pointer = 0;
      while(true)
      {
        byte[] temp = new byte[2];
        int res = in.read(temp, 0, 2);
        if (res == -1) break;

        char i = (char)((temp[1] << 8) | temp[0]);

        // Extract 5-bit ABGR;
        char alpha = (char)((i >> 15) & 0x01);
        char blue = (char)((i >> 10) & 0x1F);
        char green = (char)((i >> 5) & 0x1F);
        char red = (char)(i & 0x1F);

        // Scale the 5-bit values to 8-bit
        if (alpha == 1) alpha = 255;
        else alpha = 0;

        red = (char)((red << 3) | 0x07);
        blue = (char)((green << 3) | 0x07);
        green = (char)((blue << 3) | 0x07);

        // Combine into 24-bit value AARRGGBB
        buffer[pointer] = (alpha << 24) | (red << 16) | (green << 8) | blue;
        pointer++;
      }
      
      // img = Image.createRGBImage(buffer, img.getWidth(), img.getHeight(), true);
      img = Image.createRGBImage(buffer, 128, 64, true);
    }
    catch(Exception e)
    {
      e.printStackTrace();
      Sagiri._instance.Set_Scene(new DeathError(e));
    }
  }
  
  // TODO: handle multiline
  public short[] Measure(String str)
  {
    short left = 0;
    short right = 0;
    short top = 0;
    short bottom = 0;
    short[] cursor = new short[]{0, 0};
    boolean isFirst = true;

    for (int i=0; i<str.length(); i++)
    {
      char index = str.charAt(i);
      if (index != '\n' && (index < 32 || index > 127)) index = 127;

      if (index == '\n')
      {
        isFirst = true;
        cursor[1] += lineHeight;
        cursor[0] = 0;
        continue;
      }

      short[] ch = chars[index-32];
      int[] offset = new int[]{ch[XOFF]+cursor[0], ch[YOFF]+cursor[1]};

      // check first XOFF
      if (isFirst)
      {
        left = (short)Math.min(ch[XOFF], left);
        isFirst = false;
      }

      right = (short)Math.max(right, offset[0]+ch[WIDTH]);
      top = (short)Math.min(top, offset[1]);
      bottom = (short)Math.max(bottom, offset[1]+ch[HEIGHT]);
      
      cursor[0] += ch[XADV];
    }

    return new short[]{
      (short)(right+Math.abs(left)), // x
      (short)(bottom+Math.abs(top)), // y
      left, top // xoff, yoff
    };
  }

  public void Draw(Graphics g, String str, int x, int y)
  {
    short[] cursor = new short[]{0, 0};

    for (int i=0; i<str.length(); i++)
    {
      char index = str.charAt(i);
      if (index != '\n' && (index < 32 || index > 127)) index = 127;
      
      if (index == '\n')
      {
        cursor[1] += lineHeight;
        cursor[0] = 0;
        continue;
      }
      
      short[] ch = chars[index-32];
      int[] offset = new int[]{x+ch[XOFF]+cursor[0], y+ch[YOFF]+cursor[1]};
      g.drawRegion(
        img,
        ch[X], ch[Y], ch[WIDTH], ch[HEIGHT],
        Sprite.TRANS_NONE,
        offset[0], offset[1],
        Graphics.TOP | Graphics.LEFT);
      
      cursor[0] += ch[XADV];

      /* 
       * TODO: handle kernings here
       */
    }
  }

  public void DrawEx(Graphics g, String str, int x, int y, int anchor)
  {
    short[] rect = Measure(str);
    switch (anchor) {
      case (Graphics.TOP | Graphics.LEFT):
      Draw(g, str, x, y);
      break;

      case (Graphics.TOP | Graphics.RIGHT):
      Draw(g, str, x-rect[0], y);
      break;

      case (Graphics.BOTTOM | Graphics.LEFT):
      Draw(g, str, x, y-rect[1]);
      break;

      case (Graphics.BOTTOM | Graphics.RIGHT):
      Draw(g, str, x-rect[0], y-rect[1]);
      break;

      case (Graphics.VCENTER | Graphics.LEFT):
      Draw(g, str, x, y-(rect[1]/2));
      break;

      case (Graphics.VCENTER | Graphics.RIGHT):
      Draw(g, str, x-rect[0], y-(rect[1]/2));
      break;

      case (Graphics.TOP | Graphics.HCENTER):
      Draw(g, str, x-(rect[0]/2), y);
      break;

      case (Graphics.BOTTOM | Graphics.HCENTER):
      Draw(g, str, x-(rect[0]/2), y-rect[1]);
      break;

      case (Graphics.VCENTER | Graphics.HCENTER):
      Draw(g, str, x-(rect[0]/2), y-(rect[1]/2));
      break;

      case (Graphics.BASELINE | Graphics.LEFT):
      Draw(g, str, x, y+(lineHeight-base));
      break;

      case (Graphics.BASELINE | Graphics.RIGHT):
      Draw(g, str, x-rect[0], y+(lineHeight-base));
      break;
    
      default:
      Draw(g, str, x, y);
      break;
    }
  }

}
